import glm
import re, os, SlashBack, sys, traceback

builtin_function_or_method = type(glm.silence)

HERE = os.path.dirname(__file__)

TARGET_FOLDER = os.path.abspath(os.path.join(HERE, "../wiki/function-reference"))

functions = []

func_common = ["abs", "sign", "floor", "trunc", "round", "roundEven",
               "ceil", "fract", "mod", "min", "max", "fmin",
               "fmax", "clamp", "mix", "step", "smoothstep", "isnan",
               "isinf", "fma", "frexp", "ldexp", "floatBitsToInt",
               "floatBitsToUint", "intBitsToFloat", "uintBitsToFloat",
               "modf"]

func_exponential = ["pow", "exp", "log", "exp2", "log2", "sqrt", "inversesqrt"]

func_geometric = ["length", "distance", "dot", "cross", "normalize", "faceforward", "reflect", "refract"]

func_integer = ["uaddCarry", "usubBorrow", "umulExtended", "imulExtended",
                "bitfieldExtract", "bitfieldInsert", "bitfieldReverse",
                "bitCount", "findLSB", "findMSB"]

func_matrix = ["matrixCompMult", "outerProduct", "transpose", "determinant", "inverse"]

func_packing = ["packDouble2x32", "packUnorm2x16", "packSnorm2x16", "packSnorm4x8",
                "packUnorm4x8", "packHalf2x16", "unpackDouble2x32", "unpackUnorm2x16",
                "unpackSnorm2x16", "unpackSnorm4x8", "unpackUnorm4x8", "unpackHalf2x16"]

func_trigonometric = ["radians", "degrees", "sin", "cos", "tan", "asin", "acos", "atan",
                      "sinh", "cosh", "tanh", "asinh", "acosh", "atanh"]

func_vector_relational = ["equal", "notEqual", "lessThan", "lessThanEqual", "greaterThan",
                          "greaterThanEqual", "any", "all", "not_"]

detail = func_common + func_exponential + func_geometric + func_integer + func_matrix + \
         func_packing + func_trigonometric + func_vector_relational

color_space = ["convertLinearToSRGB", "convertSRGBToLinear"]

constants = ["epsilon", "zero", "one", "pi", "two_pi", "root_pi",
             "half_pi", "three_over_two_pi", "quarter_pi", "one_over_pi",
             "one_over_two_pi", "two_over_pi", "four_over_pi",
             "two_over_root_pi", "one_over_root_two", "root_half_pi",
             "root_two_pi", "root_ln_four", "e", "euler", "root_two",
             "root_three", "root_five", "ln_two", "ln_ln_two", "ln_ten", "third",
             "two_thirds", "golden_ratio"]

epsilon = ["epsilonEqual", "epsilonNotEqual"]

integer = ["iround", "uround"]

matrix_access = ["row", "column"]

matrix_clip_space = ["tweakedInfinitePerspective"] + \
                    [x + y for y in ["", "LH", "RH", "NO", "ZO", "LH_ZO", "LH_NO", "RH_ZO", "RH_NO"] for x in ["ortho", "frustum", "perspective", "perspectiveFov"]] +\
                    ["infinitePerspective" + y for y in ["", "LH", "RH"]]

matrix_inverse = ["affineInverse", "inverseTranspose"]

matrix_projection = ["project", "projectNO", "projectZO", "unProject", "unProjectNO", "unProjectZO", "pickMatrix"]

matrix_transform = ["identity", "translate", "rotate", "rotate_slow", "scale", "scale_slow", "lookAt",
                    "lookAtRH", "lookAtLH"]

noise = ["perlin", "simplex"]

packing_pack_methods = ["pack" + x + y for x, z in [
    ("Int", ("2x8", "4x8", "4x16", "2x16", "2x32")),
    ("Uint", ("2x8", "4x8", "4x16", "2x16", "2x32")),
    ("Unorm", ("", "1x8", "2x8", "1x16", "4x16", "3x10_1x2", "2x4", "4x4", "1x5_1x6_1x5", "3x5_1x1", "2x3_1x2")),
    ("Snorm", ("", "1x8", "2x8", "1x16", "4x16", "3x10_1x2")),
    ("Half", ("", "1x16", "4x16")),
    ("", ("I3x10_1x2", "U3x10_1x2", "F2x11_1x10", "F3x9_E1x5", "RGBM")),
    ] for y in z]

packing = [y + x for x in packing_pack_methods for y in ["", "un"]]

quaternion = ["quatLookAt", "quatLookAtLH", "quatLookAtRH", "eulerAngles", "roll", "pitch", "yaw", "mat3_cast",
              "mat4_cast", "quat_cast"]

quaternion_common = ["lerp", "slerp", "conjugate"]

quaternion_trigonometric = ["angle", "axis", "angleAxis"]

random = ["linearRand", "gaussRand", "circularRand", "sphericalRand", "diskRand", "ballRand", "setSeed"]

reciprocal = [f.format(x) for f in ("{}", "a{}", "{}h", "a{}h") for x in ("sec", "csc", "cot")]

round = ["ceilPowerOfTwo", "floorPowerOfTwo", "roundPowerOfTwo", "ceilMultiple", "floorMultiple",
         "roundMultiple"]

type_ptr = ["value_ptr", "sizeof", "make_quat"] + [f + str(x) for x in range(2, 5) for f in ("make_vec", "make_mat")] +\
           ["make_mat{}x{}".format(x, y) for x in range(2,5) for y in range(2,5)]

ulp = ["next_float", "prev_float", "float_distance"]

decompose = ["decompose"]

norm = ["distance2", "l1Norm", "l2Norm", "lMaxNorm", "length2", "lxNorm"]

polar_coordinates = ["polar", "euclidean"]

matrix_transform_2d = ["shearX", "shearY", "rotate", "translate", "scale"]

rotate_vector = ["orientation", "rotate", "rotateX", "rotateY", "rotateZ", "slerp"]

compatibility = ["lerp", "atan2", "isfinite", "saturate"]

stable_extensions = matrix_clip_space + matrix_projection + matrix_transform + quaternion_common +\
                    quaternion_trigonometric

recommended_extensions = color_space + constants + epsilon + integer + matrix_access + matrix_inverse +\
                         noise + packing + quaternion + random + reciprocal + round + type_ptr + ulp

unstable_extensions = decompose + norm + polar_coordinates + matrix_transform_2d + rotate_vector + compatibility

other_functions = ["silence", "add", "and_", "or_", "xor", "cmp", "div", "floordiv", "if_else", "inv", "lshift", "mul", "neg", "pos", "rshift", "sub", "quat_to_vec4", "vec4_to_quat"]

known_functions = detail + recommended_extensions + stable_extensions + unstable_extensions + other_functions

ignored_functions = ["test", "test_type_identification"]

multiple_expected = ["rotate", "translate", "scale", "slerp", "lerp"]

folders = {}

for raw in dir(glm):
    func = getattr(glm, raw)
    if (raw.startswith("__") and raw.endswith("__") or not isinstance(func, builtin_function_or_method)):
        continue
    functions.append(raw)

is_defined = lambda x: x in functions
is_known = lambda x: x in known_functions
is_ignored = lambda x: x in ignored_functions

for known_func in known_functions:
    if not is_defined(known_func):
        print("A 'known' function was not defined: {}".format(known_func))

    if known_func not in multiple_expected and known_functions.count(known_func) != 1:
        print("Function '{}' is known multiple times".format(known_func))


for function in functions:
    if not is_known(function) and not is_ignored(function):
        print("Function {} is not known".format(function))

FUNC_PATTERN = re.compile("(^\\w+)\\s*\\(([^)]*)\\)\\s*->\\s*(.+$)")

TEMPLATE_PATTERN = re.compile("\\\\code\\\\([A-Z]) in \\[((:?\\w+,? ?)+)\\]\\\\code\\\\")

class FunctionName:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "glm.<code>**{}**".format(self.name)

class Parameters:
    def __init__(self, param_string):
        self.param_string = param_string

    def __str__(self):
        inner_out = []
        for param in self.param_string.split(","):
            param = param.strip()
            if not param:
                continue
            if ":" in param:
                if param.count(":") != 1:
                    print("Invalid amount of ':': {}".format(self.param_string))
                else:
                    param, type_ = param.split(":")
                    inner_out.append("**{}**: *{}*".format(param, type_.strip()))
            else:
                inner_out.append("**{}**".format(param))
                
        return "({})".format(", ".join(inner_out))

class ReturnValue:
    def __init__(self, return_value):
        self.return_value = return_value

    def __str__(self):
        return " -\\\\> *{}*".format(self.return_value)

class Description:
    def __init__(self, descr):
        self.descr = descr

    def __str__(self):
        descr = self.descr.replace("\t", "\\raw\\&emsp;&emsp;\\raw\\")
        while 1:
            match = TEMPLATE_PATTERN.search(descr)
            if match:
                descr = descr[:match.start()] + "\\raw\\<code>***{}*** in [{}]</code>\\raw\\".format(match.group(1), ", ".join(map(lambda x: "**{}**".format(x), match.group(2).split(", ")))) + descr[match.end():]
            else:
                break
        return descr


def convert_doc_entry_to_sb(subdoc):
    lines = subdoc.split("\n")
    first_line = ""
    rest_start = 0
    for line in lines:
        first_line += line
        rest_start += 1
        if "->" in line:
            break
    descr = Description("\n".join(lines[rest_start:]))
    match = FUNC_PATTERN.fullmatch(first_line)
    if match:
        func_name = FunctionName(match.group(1))
        parameters = Parameters(match.group(2))
        return_val = ReturnValue(match.group(3))
        return "\\raw\\#### <code>{}{}{}</code></code>\\raw\\\n{}\n".format(str(func_name), str(parameters), str(return_val), str(descr))
    print("{} doesn't match the FUNC_PATTERN!".format(subdoc))
    return ""

def convert_doc_to_sb(func):
    doc = func.__doc__
    name = func.__name__
    if doc.count("`") % 2 != 0:
        print ("{} doesn't have an even amount of backticks!".format(name))
    else:
        doc = doc.replace("`", "\\code\\")
    out = "\\h3\\{}() function\\h3\\\n".format(name)
    lines = doc.split("\n")
    current_subdoc = ""
    for line in lines:
        if not line.startswith("\t") and current_subdoc:
            out += convert_doc_entry_to_sb(current_subdoc)
            current_subdoc = ""
        current_subdoc += line + "\n"
    if current_subdoc.strip():
        out += convert_doc_entry_to_sb(current_subdoc)
    return out

def check_doc(func):
    name = func.__name__
    doc = func.__doc__
    for line in doc.split("\n"):
        if line.strip() and not line.startswith("\t") and not line.startswith(name) or len(line) > 100:
            print("{} has invalid doc: \n{}\n".format(name, line))
            return False
    return True

def convert_sb_file(filename):
    out_filename = os.path.splitext(filename)[0] + ".md"
    try:
        sbf = SlashBack.open(filename)
        sbf.write(out_filename)
    except SlashBack.SlashBackParserError as exception:
        print(f"\nERROR: {exception!s}", file=sys.stderr)
    except Exception as exception:
        print("\nUNHANDLED EXCEPTION OCCURED!", file=sys.stderr)
        traceback.print_exc()

def generate_reference_for(name, function_names, description):
    folder = "detail" if function_names[0] in detail else "stable_extensions" if function_names[0] in stable_extensions else \
        "recommended_extensions" if function_names[0] in recommended_extensions else "unstable_extensions" if function_names[0] in unstable_extensions else \
        "other"

    if not folder in folders:
        folders[folder] = {}

    folders[folder][name] = function_names

    if not os.path.exists(TARGET_FOLDER + "/{}/".format(folder)):
        os.mkdir(TARGET_FOLDER + "/{}/".format(folder))

    file_name = TARGET_FOLDER + "/{}/{}.sb".format(folder, name)

    file = open(file_name, "w")
    file.write("\\c\\This file was generated using a tool\\c\\\n")
    file.write("\\h1\\{} methods\\h1\\\n".format(name))
    file.write("The following methods are all part of the \\b\\{} methods\\b\\.\n".format(name))
    file.write(description)
    file.write("\n\\h2\\Table of contents\\h2\\\n")
    file.write("\\ul\\\n")

    function_names.sort()
    
    for function_name in function_names:
        file.write("\\-\\\\url #{}-function\\\\b\\{}\\b\\ function\\url\\\n".format(function_name.lower(), function_name))
    file.write("\\ul\\\n")
    
    for function_name in function_names:
        if is_defined(function_name):
            func = getattr(glm, function_name)
            if check_doc(func):
                file.write(convert_doc_to_sb(func))
    file.close()
    convert_sb_file(file_name)

        
generate_reference_for("func_common", func_common, "It contains common GLSL functions.")

generate_reference_for("func_exponential", func_exponential, "It contains exponential GLSL functions.")

generate_reference_for("func_geometric", func_geometric, "These operate on vectors as vectors, not component-wise.")

generate_reference_for("func_integer", func_integer, "It contains GLSL functions on integer types.")

generate_reference_for("func_matrix", func_matrix, "It contains GLSL matrix functions.")

generate_reference_for("func_packing", func_packing, "Provides GLSL functions to pack and unpack half, single"+\
                       " and double-precision floating point values into more compact integer types.")

generate_reference_for("func_trigonometric", func_trigonometric,
                       "Function parameters specified as angle are assumed to be in units of radians.")

generate_reference_for("func_vector_relational", func_vector_relational,
                       "Relational vector functions.")

generate_reference_for("color_space", color_space, "Allow to perform bit operations on integer values.")

generate_reference_for("constants", constants, "Provide a list of constants and precomputed useful values.")

generate_reference_for("epsilon", epsilon, "Comparison functions for a user defined epsilon values.")

generate_reference_for("integer", integer, "Contains two different rounding methods.")

generate_reference_for("matrix_access", matrix_access,
                       "Defines functions to access rows or columns of a matrix easily.")

generate_reference_for("matrix_clip_space", matrix_clip_space,
                       "Defines functions that generate clip space transformation matrices.")

generate_reference_for("matrix_inverse", matrix_inverse,
                       "Defines additional matrix inverting functions.")

generate_reference_for("matrix_projection", matrix_projection,
                       "Functions that generate common projection transformation matrices.")

generate_reference_for("matrix_transform", matrix_transform,
                       "Defines functions that generate common transformation matrices.")

generate_reference_for("noise", noise,
                       "Defines 2D, 3D and 4D procedural noise functions Based on the work of Stefan Gustavson"+\
                       " and Ashima Arts on \"webgl-noise\": \\raw\\https://github.com/ashima/webgl-noise\\raw\\ Following"+\
                       " Stefan Gustavson's paper \"Simplex noise demystified\": "+\
                       "\\raw\\http://www.itn.liu.se/~stegu/simplexnoise/simplexnoise.pdf\\raw\\")

generate_reference_for("packing", packing,
                       "This extension provides a set of function to convert vertors to packed formats.")

generate_reference_for("quaternion", quaternion,
                       "Defines a templated quaternion type and several quaternion operations.")

generate_reference_for("quaternion_common", quaternion_common,
                       "Provides common functions for quaternion types.")

generate_reference_for("quaternion_trigonometric", quaternion_trigonometric,
                       "Provides trigonometric functions for quaternion types.")

generate_reference_for("random", random,
                       "Generate random number from various distribution methods.")

generate_reference_for("reciprocal", reciprocal,
                       "Define secant, cosecant and cotangent functions.")

generate_reference_for("round", round,
                       "Rounding value to specific boundings.")

generate_reference_for("type_ptr", type_ptr,
                       "Handles the interaction between pointers and vector, matrix types. "+\
                       "This extension defines an overloaded function, glm.value_ptr. It "+\
                       "returns a pointer to the memory layout of the object. Matrix types "+\
                       "store their values in column-major order. This is useful for uploading "+\
                       "data to matrices or copying data to buffer objects.")

generate_reference_for("ulp", ulp,
                       "Allow the measurement of the accuracy of a function against a reference implementation. "+\
                       "This extension works on floating-point data and provide results in ULP.")

generate_reference_for("decompose", decompose,
                       "Decomposes a model matrix to translations, rotation and scale components.")

generate_reference_for("norm", norm,
                       "Various ways to compute vector norms.")

generate_reference_for("polar_coordinates", polar_coordinates,
                       "Conversion from Euclidean space to polar space and revert.")

generate_reference_for("matrix_transform_2d", matrix_transform_2d,
                       "Defines functions that generate common 2d transformation matrices.")

generate_reference_for("rotate_vector", rotate_vector,
                       "Function to directly rotate a vector.")

generate_reference_for("compatibility", compatibility,
                       "Provide functions to increase the compatibility with Cg and HLSL languages.")

generate_reference_for("other", other_functions,
                       "PyGLM's custom functions.")

instructions_filename = TARGET_FOLDER + "/INSTRUCTIONS.sb"

instructions_file = open(instructions_filename, "r")
instuctions = instructions_file.read()
instructions_file.close()

toc_filename = TARGET_FOLDER + "/README.sb"

toc_file = open(toc_filename, "w")
toc_file.write("{}\n".format(instuctions))
toc_file.write("\\h1\\Table of Contents\\h1\\\n")
toc_file.write("\\ul\\\n")

for folder in folders:
    sub_toc_filename = TARGET_FOLDER + "/{}/README.sb".format(folder)
    sub_toc_file = open(sub_toc_filename, "w")
    sub_toc_file.write("\\h1\\Table of Contents\\h1\\\n")
    sub_toc_file.write("\\ul\\\n")
    toc_file.write("\\-\\ \\url {0}/README.md\\{0}\\url\\\n".format(folder))
    folder_content = folders[folder]
    for functions_name in folder_content:
        sub_toc_file.write("\\-\\ \\url {0}.md \\{0} methods\\ url\\\n".format(functions_name))
        toc_file.write("\\--\\ \\url {0}/{1}.md \\{1} methods\\ url\\\n".format(folder, functions_name))
        functions = folder_content[functions_name]
        for function_name in functions:
            sub_toc_file.write("\\--\\ \\url {0}.md#{1}-function \\\\b\\{1}\\b\\ function \\ url\\\n".format(functions_name, function_name))
            toc_file.write("\\---\\ \\url {0}/{1}.md#{2}-function \\\\b\\{2}\\b\\ function \\ url\\\n".format(folder, functions_name, function_name))

    sub_toc_file.write("\\ul\\\n")
    sub_toc_file.close()
    convert_sb_file(sub_toc_filename)

toc_file.write("\\ul\\\n")

toc_file.close()

convert_sb_file(toc_filename)
























import os

os.environ["TEST_PICKLING"] = "True"

TEST = True

TEST_PERFORMANCE = False

GEN_DOC = True

if TEST_PERFORMANCE:
    import PyGLM_vs_NumPy

if TEST:
    import PyGLM_test, time, traceback, sys
    print("Testing PyGLM...")
    start_time = time.time()
    for member in dir(PyGLM_test):
        if member.startswith("test"):
            try:
                getattr(PyGLM_test, member)()
            except:
                print("The following test failed: " + member, file=sys.stderr)
                traceback.print_exc()
    print("Finished tests in {:.3g}s".format(time.time()-start_time))


import glm
from glm import *

if GEN_DOC:
    import generate_function_reference

if True:
    import time
    def t(func, *args, r=1000000, **kw):
        start = time.time()
        for i in range(r):
            func(*args, **kw)
        return time.time() - start
    from sys import getrefcount as rc
    def memcor(func, *args, r=1000000, **kw):
        for i in range(r):
            try:
                func(*args, **kw)
            except:
                pass





from pyglm.glm import *
from pyglm import glm_typing, typing, __version__

import warnings as __warnings

__warnings.warn("Importing PyGLM via \"import glm\" is going to be deprecated in the future.\nUse \"from pyglm import glm\" instead", PendingDeprecationWarning, stacklevel=2)



from setuptools import Extension, setup

from codecs import open

from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "VERSION")) as file_:
    VERSION = file_.read().strip()

setup(
    ext_modules=[
        Extension(
            name="pyglm.glm",
            sources=["PyGLM_lib/PyGLM.cpp"],
            include_dirs=["PyGLM_lib/glm/"], 
            extra_compile_args=['-std=c++11', f'-DVERSION={VERSION}']
        ),
    ]
)



import timeit, sys, random
import glm, numpy

ON_WINDOWS = sys.platform == 'win32'

print_horizontal_rule = lambda: print("+----------------------------------------+------------+------------+-----------+")

pad_with_spaces = lambda text, length, align="left": text + " " * (length - len(text)) if align == "left" else " " * (length - len(text)) + text

seconds_to_milliseconds = lambda seconds: int(round(seconds * 1000, 0))

pyglm_total_time = 0
numpy_total_time = 0

def word_wrap(text, max_length):
    out = []
    current_word = []
    current_length = 0
    for c in text:
        if current_length >= max_length:
            out.append("\n")
            current_length = len(current_word)

        if c == " ":
            out.append("".join(current_word))
            if (current_length < max_length):
                out.append(" ")
            current_word = []
        elif c == "\n":
            out.append("".join(current_word))
            out.append("\n")
            current_word = []
            current_length = 0
        else:
            current_word.append(c)

        current_length += 1

    if current_word:
        out.append("".join(current_word))

    return "".join(out).rstrip()

def print_row(descr, pyglm, numpy, ratio, print_header=True, end="\n"):
    descr = word_wrap(descr, 38)

    descr_lines = descr.split("\n")

    if print_header:
        for line in descr_lines[:-1]:
            print(f"| {pad_with_spaces(line, 38)} | {' ' * 10} | {' ' * 10} | {' ' * 9} |")

    descr_last_line = descr_lines[-1]
    
    print(f"| {pad_with_spaces(descr_last_line, 38)} | {pad_with_spaces(pyglm, 10, align='right')} | {pad_with_spaces(numpy, 10, align='right')} | {pad_with_spaces(ratio, 9, align='right')} |", end=end)

def run_test(descr, pyglm_setup_code, pyglm_code, numpy_setup_code, numpy_code, number):
    global pyglm_total_time, numpy_total_time

    descr += f"\n({number:,} times)"
    
    if ON_WINDOWS:
        print_row(descr, "", "", "", end="\r")
    
    pyglm_result = 2**31
    numpy_result = 2**31

    for i in range(10):
        run_pyglm_first = random.choice((True,False))

        if run_pyglm_first:
            pyglm_result = min(pyglm_result, timeit.timeit(pyglm_code, pyglm_setup_code, number=number))
            numpy_result = min(numpy_result, timeit.timeit(numpy_code, numpy_setup_code, number=number))
        else:
            numpy_result = min(numpy_result, timeit.timeit(numpy_code, numpy_setup_code, number=number))
            pyglm_result = min(pyglm_result, timeit.timeit(pyglm_code, pyglm_setup_code, number=number))
            
    
        if ON_WINDOWS:
            print_row(descr, "{}ms".format(seconds_to_milliseconds(pyglm_result)), "{}ms".format(seconds_to_milliseconds(numpy_result)),  "{:.02f}x".format(numpy_result / pyglm_result), end="\r", print_header=False)

    pyglm_total_time += pyglm_result
    numpy_total_time += numpy_result
    
    print_row(descr, "{}ms".format(seconds_to_milliseconds(pyglm_result)), "{}ms".format(seconds_to_milliseconds(numpy_result)),  "{:.02f}x".format(numpy_result / pyglm_result), print_header=not ON_WINDOWS)
    print_horizontal_rule()

print(f"""Evaluating performance of PyGLM compared to NumPy.

Running on platform '{sys.platform}'.

Python version:
{sys.version}

Comparing the following module versions:
{glm.version}
 vs
NumPy version {numpy.__version__}
________________________________________________________________________________

The following table shows information about a task to be achieved and the time
it took when using the given module. Lower time is better.
Each task is repeated ten times per module, only showing the best (i.e. lowest)
value.

""")

print_horizontal_rule()
print_row("Description", "PyGLM time", "NumPy time", "ratio")
print_horizontal_rule()

############################
# Actual tests start here: #
############################

run_test("3 component vector creation",
         
         "import glm",
         "glm.vec3()",

         "import numpy",
         "numpy.zeros((3,), numpy.float32)",

         100000
)

run_test("3 component vector creation with custom components",
         
         "import glm",
         "glm.vec3(1,2,3)",

         "import numpy",
         "numpy.array((1,2,3), numpy.float32)",

         50000
)

run_test("dot product",
         
         "import glm; v1 = glm.vec3(); v2 = glm.vec3()",
         "glm.dot(v1, v2)",
         
         "import numpy; v1 = numpy.zeros((3,), numpy.float32); v2 = numpy.zeros((3,), numpy.float32)",
         "numpy.dot(v1, v2)",

         50000
)

run_test("cross product",
         
         "import glm; v1 = glm.vec3(1); v2 = glm.vec3(1,2,3)",
         "glm.cross(v1, v2)",

         "import numpy; v1 = numpy.array((1,1,1), numpy.float32); v2 = numpy.array((1,2,3), numpy.float32)",
         "numpy.cross(v1, v2)",

         25000
)

run_test("L2-Norm of 3 component vector",
         
         "import glm; v = glm.vec3(1,2,3)",
         "glm.l2Norm(v)",
         
         "import numpy; v = numpy.array((1,1,1), numpy.float32)",
         "numpy.linalg.norm(v)",

         100000
)

run_test("4x4 matrix creation",
         
         "import glm",
         "glm.mat4(0)",
         
         "import numpy",
         "numpy.zeros((4,4), numpy.float32)",

         50000
)
run_test("4x4 identity matrix creation",
         
         "import glm",
         "glm.mat4()",
         
         "import numpy",
         "numpy.identity(4, numpy.float32)",

         100000
)

run_test("4x4 matrix transposition",
         
         "import glm; m = glm.mat4()",
         "glm.transpose(m)",
         
         "import numpy; m = numpy.identity(4, numpy.float32)",
         "numpy.transpose(m)",

         50000
)

run_test("4x4 multiplicative inverse",
         
         "import glm; m = glm.mat4()",
         "glm.inverse(m)",
         
         "import numpy; m = numpy.identity(4, numpy.float32)",
         "numpy.linalg.inv(m)",

         50000
)

run_test("3 component vector addition",
         
         "import glm; v1 = glm.vec3(1); v2 = glm.vec3(1,2,3)",
         "v1 + v2",
         
         "import numpy; v1 = numpy.array((1,1,1), numpy.float32); v2 = numpy.array((1,2,3), numpy.float32)",
         "v1 + v2",

         100000
)

run_test("4x4 matrix multiplication",
         
         "import glm; m1 = glm.mat4(); m2 = glm.mat4(2)",
         "m1 * m2",
         
         "import numpy; m1 = numpy.identity(4, numpy.float32); m2 = numpy.identity(4, numpy.float32) * 2",
         "m1 * m2",

         100000
)

run_test("4x4 matrix x vector multiplication",
         
         "import glm; m = glm.mat4(); v = glm.vec4()",
         "m * v",
         
         "import numpy; m = numpy.identity(4, numpy.float32); v = numpy.zeros((4,), numpy.float32)",
         "m * v",

         100000
)

print_row("TOTAL",
          "{:.02f}s".format(pyglm_total_time),
          "{:.02f}s".format(numpy_total_time),
          "{:.02f}x".format(numpy_total_time / pyglm_total_time))
print_horizontal_rule()




# License (zlib/libpng):

#############################################
#                                            #
#  Copyright (c) 2019 Zuzu_Typ               #
#--------------------------------------------#
#  This software is provided 'as-is',        #
#  without any express or implied warranty.  #
#  In no event will the authors be held      #
#  liable for any damages arising from the   #
#  use of this software.                     #
#                                             #
#  Permission is granted to anyone to use     #
#  this software for any purpose, including   #
#  commercial applications, and to alter it   #
##  and redistribute it freely, subject to     #
##  the following restrictions:                #
##                                             #
##  1. The origin of this software must not    #
##     be misrepresented; you must not claim   #
##     that you wrote the original software.   #
##     If you use this software in a product,  #
##     an acknowledgment in the product        #
##     documentation would be appreciated but  #
##     is not required.                        #
##                                             #
##  2. Altered source versions must be         #
# #     plainly marked as such, and must not    #
# #     be misrepresented as being the          #
# #     original software.                      #
# #                                             #
# #  3. This notice may not be removed or       #
###     altered from any source distribution.   #
  ##############################################

from builtins import open as builtins_open

import sys, os, traceback, string

__version__ = "0.2.0"

__welcome__ = f"""\
SlashBack v{__version__} by Zuzu_Typ"""

__doc__ = """\
A little editor for my own markup language
called SlashBack ( \\SB\\ )."""

__usage__ = """\
Usage:
SlashBack.py <input_file(s)>"""

class TextConfig:
    NO_LIST         = 0
    ORDERED_LIST    = 1
    UNORDERED_LIST  = 2

    NO_TASK         = 0
    CHECKED_TASK    = 1
    UNCHECKED_TASK  = 2
    
    bold            = False
    italic          = False
    strikethrough   = False
    underline       = False
    quote           = False
    table           = False
    separator       = False
    raw             = False
    ignore          = False
    code            = None
    image           = None
    url             = None
    mention         = None
    reference       = None
    comment         = None
    list            = NO_LIST
    task            = NO_TASK
    header          = 0
    index           = 1
    indent          = 1
    list_indices    = {} # only for ordered lists
    
    def __init__(self, other_config = None):
        if other_config:
            self.bold           = other_config.bold
            self.italic         = other_config.italic
            self.strikethrough  = other_config.strikethrough
            self.underline      = other_config.underline
            self.quote          = other_config.quote
            self.table          = other_config.table
            self.separator      = other_config.separator
            self.raw            = other_config.raw
            self.ignore         = other_config.ignore
            self.code           = other_config.code
            self.image          = other_config.image
            self.url            = other_config.url
            self.mention        = other_config.mention
            self.reference      = other_config.reference
            self.comment        = other_config.comment
            self.list           = other_config.list
            self.task           = other_config.task
            self.header         = other_config.header
            self.index          = other_config.index
            self.indent         = other_config.indent
            self.list_indices   = other_config.list_indices.copy()

    def __str__(self):
        out = set()

        if self.comment:
            out.add(f"/* {self.comment} */")
        if self.bold:
            out.add("bold")
        if self.italic:
            out.add("italic")
        if self.strikethrough:
            out.add("strikethrough")
        if self.underline:
            out.add("underline")
        if self.quote:
            out.add("quote")
        if self.table:
            out.add(f"table at {self.index}")
        if self.raw:
            out.add("<raw>")
        if self.code:
            out.add(f"code {self.code}")
        if self.image:
            out.add(f"image {self.image}")
        if self.url:
            out.add(f"url {self.url}")
        if self.mention:
            out.add(f"mention {self.mention}")
        if self.reference:
            out.add(f"reference {self.reference}")
        if self.list:
            out.add(("ordered list" if self.list == self.ORDERED_LIST else "unordered list") + f" at {self.index}")
        if self.task:
            out.add("checked task" if self.list == self.CHECKED_TASK else "unchecked task")
        if self.header:
            out.add(f"h{self.header}")
        if self.separator:
            out.add("---")

        return ", ".join(out)

    __repr__ = lambda self: self.__str__()

class Text:
    def __init__(self, text, text_config):
        self.text = text
        self.text_config = text_config

class ParsedText:
    def __init__(self, text_list):
        self.text_list = text_list

    def to_markdown(self):
        out = [f"[//]: # (generated using SlashBack {__version__})\n\n".encode()]

        current_text_config = TextConfig()

        table_signalized = True

        for text in self.text_list:
            text_string = text.text
            this_text_config = text.text_config

            if this_text_config.raw:
                out.append(text_string)
                continue

            if not this_text_config.code:
                for char in "\\`*_{}[]()#+-.!":
                    text_string = text_string.replace(char.encode(), f"\\{char}".encode())
                for char, replacement in ((b"&", b"&amp;"), (b"<", b"&lt;"), (b">", b"&gt;"), (b"~", b"&#126;"), (b"|", b"&#124;"), (b"@", b"@<!---->"), (b"`", b"&#96;")):
                    text_string = text_string.replace(char, replacement)

                if not this_text_config.table:
                    text_string = text_string.replace(b"\t", b"    ").replace(b"  ", b"&nbsp;" * 2).replace(b"\r\n", b"\n").replace(b"\r", b"\n").replace(b"\n", b"  \n")
            else:
                pass#text_string = text_string.replace(b"`", b"&#96;")

            if not table_signalized and b"\n" in text_string:
                table_signalized = True
                text_string = text_string.replace(b"\n", b"\n" + b"|".join([b"-"] * (this_text_config.index + 1)) + b"\n")

            

            if this_text_config.list and this_text_config.index != 0 and current_text_config.index != this_text_config.index:
                out.append(b"    " * (this_text_config.indent - 1) + ( f"{this_text_config.index}. ".encode() if this_text_config.list == TextConfig.ORDERED_LIST else b"* "))

            if current_text_config.table and this_text_config.table and current_text_config.index < this_text_config.index:
                out.append(b"|")

            if not current_text_config.bold and this_text_config.bold:
                out.append(b"**")

            if not current_text_config.italic and this_text_config.italic:
                out.append(b"*")

            if not current_text_config.strikethrough and this_text_config.strikethrough:
                out.append(b"~~")

            if not current_text_config.underline and this_text_config.underline:
                out.append(b"<u>")

            if not current_text_config.quote and this_text_config.quote:
                out.append(b">")

            if not current_text_config.table and this_text_config.table:
                table_signalized = False

            if not current_text_config.code and this_text_config.code:
                out.append(b"``` ")
                if isinstance(this_text_config.code, bytes):
                    out.append(this_text_config.code)

            if not current_text_config.image and this_text_config.image:
                out.append(b"![")

            if not current_text_config.url and this_text_config.url:
                out.append(b"[")

            if not current_text_config.list and this_text_config.list:
                pass

            if not current_text_config.task and this_text_config.task:
                if this_text_config.task == TextConfig.CHECKED_TASK:
                    out.append(b"- [x] ")
                else:
                    out.append(b"- [ ] ")

            if not current_text_config.header and this_text_config.header:
                out.append(b"#" * this_text_config.header + b" ")

            if not current_text_config.separator and this_text_config.separator:
                out.append(b"\n---")

            if not current_text_config.mention and this_text_config.mention:
                out.append(b"@" + this_text_config.mention)

            if not current_text_config.reference and this_text_config.reference:
                out.append(this_text_config.reference)

            if not current_text_config.comment and this_text_config.comment:
                out.append(f" [//]: # ({this_text_config.comment.decode()}) ".encode())
            
            # backward #

            if current_text_config.url and not this_text_config.url:
                out.append(f"]({current_text_config.url.decode()})".encode())

            if current_text_config.image and not this_text_config.image:
                out.append(f"]({current_text_config.image.decode()})".encode())

            if current_text_config.code and not this_text_config.code:
                out.append(b" ```")

            if current_text_config.task and not this_text_config.task:
                out.append(b"\n\n")

            if current_text_config.quote and not this_text_config.quote:
                out.append(b"\n\n")

            if current_text_config.underline and not this_text_config.underline:
                out.append(b"</u>")

            if current_text_config.strikethrough and not this_text_config.strikethrough:
                out.append(b"~~")
                
            if current_text_config.italic and not this_text_config.italic:
                out.append(b"*")
                
            if current_text_config.bold and not this_text_config.bold:
                out.append(b"**")

            if not (not current_text_config.ignore and this_text_config.ignore):
                out.append(text_string)

            current_text_config = this_text_config

        if current_text_config.url:
            out.append(f"]({current_text_config.url.decode()})".encode())

        if current_text_config.image:
            out.append(f"]({current_text_config.image.decode()})".encode())

        if current_text_config.code:
            out.append(b" ```")

        if current_text_config.quote:
            out.append(b"\n\n")

        if current_text_config.strikethrough:
            out.append(b"~~")
            
        if current_text_config.italic:
            out.append(b"*")
            
        if current_text_config.bold:
            out.append(b"**")
            
        return b"".join(out)
                

class SlashBackParserError(Exception):
    pass

class SlashBackParser:
    def __init__(self, path = None, auto_read = True):
        self.path = path
        self.read()

    def read(self):
        file_ = builtins_open(self.path, "rb")
        content = file_.read()
        file_.close()

        self.parsed_text = self.parse(content)

    def write(self, path):
        file_ = builtins_open(path, "wb")
        file_.write(self.parsed_text.to_markdown())
        file_.close()

    def parse_single(self, command, config):        
        command = command.lstrip((string.whitespace + "!#/:|+.~>_<").encode()).rstrip()

        if config.separator: config.separator   = False
        if config.reference: config.reference   = None
        if config.mention: config.mention       = None
        if config.comment: config.comment       = None
        

        if command == b"raw":
            config.raw = not config.raw
            
        elif command == b"h1":
            config.header = 0 if config.header else 1
        elif command == b"h2":
            config.header = 0 if config.header else 2
        elif command == b"h3":
            config.header = 0 if config.header else 3
        elif command == b"h4":
            config.header = 0 if config.header else 4
        elif command == b"h5":
            config.header = 0 if config.header else 5
        elif command == b"h6":
            config.header = 0 if config.header else 6

        elif command == b"b":
            config.bold = not config.bold
        elif command == b"i":
            config.italic = not config.italic

        elif command in (b"separator", b"sep"):
            config.separator = True

        elif command == b"s":
            config.strikethrough = not config.strikethrough

        elif command == b"u":
            config.underline = not config.underline

        elif command in (b"quote", b"q"):
            config.quote = not config.quote

        elif command in (b"table", b"tbl"):
            if config.list:
                raise SlashBackParserError("A table can't be created inside a list")
            config.table = not config.table
            config.index = 0

        elif command in (b"list switch", b"ls", b"sl"):
            if not config.list:
                raise SlashBackParserError("There's no list to switch here")
            config.list = config.ORDERED_LIST if config.list == config.UNORDERED_LIST else config.UNORDERED_LIST
            config.index = 0

        elif command in (b"list ordered", b"lo", b"ol"):
            if config.table:
                raise SlashBackParserError("A list can't be created inside a table")
            config.list = config.NO_LIST if config.list else config.ORDERED_LIST
            config.index = 0

        elif command in (b"list unordered", b"lu", b"ul"):
            if config.table:
                raise SlashBackParserError("A list can't be created inside a table")
            config.list = config.NO_LIST if config.list else config.UNORDERED_LIST
            config.index = 0

        elif command in (b"list", b"lo", b"ol", b"lu", b"ul"):
            if not config.list:
                raise SlashBackParseError("Unspecified list type")
            
            config.list = config.NO_LIST

            config.list_indices = {}

        elif command in (b"task checked", b"tc", b"x"):
            config.task = config.NO_TASK if config.task else config.CHECKED_TASK

        elif command in (b"task unchecked", b"tu", b"o"):
            config.task = config.NO_TASK if config.task else config.UNCHECKED_TASK

        elif command in (b"task", b"tc", b"tu", b"x", b"o"):
            if not config.task:
                raise SlashBackParseError("Unspecified task type")
            
            config.task = config.NO_TASK

        elif command in (b"image", b"img"):
            if not config.image:
                raise SlashBackParseError("Unspecified image type")

            config.image = None

        elif command == b"url": #TODO
            if not config.url:
                raise SlashBackParseError("Unspecified url link")

            config.url = None

        elif command.count(b"-") == len(command):
            if config.list:
                config.indent = command.count(b"-")
                
                for indent in config.list_indices.copy():
                    if indent > config.indent:
                        del config.list_indices[indent]
                        
                if config.list == config.ORDERED_LIST:
                    config.index = config.list_indices.get(config.indent, 1)
                    
                    config.list_indices[config.indent] = config.index + 1
                else:
                    config.index += 1
                    
            elif config.table:
                config.index += 1
            if not config.list and not config.table:
                raise SlashBackParserError(f"Unexpected \\{command.decode()}\\ token")

        else:
            space_index = command.find(b" ")
            first_command = command[:space_index] if space_index != -1 else command
            info = command[space_index:].strip() if space_index != -1 else ""

            if first_command == b"image":
                config.image = None if config.image else info

            elif first_command == b"url":
                config.url = None if config.url else info

            elif first_command == b"code":
                config.code = None if config.code else info if info else True

            elif first_command == b"mention":
                config.mention = info

            elif first_command == b"reference":
                config.reference = info

            elif first_command in (b"c", b"comment", b"ignore"):
                config.ignore = not config.ignore

            elif first_command in (b"ic", b"icomment"):
                config.comment = info

            else:
                raise SlashBackParserError(f"Unknown command \"{first_command.decode()}\"")
            
        return config

    def parse(self, text):
        if (text.count(b"\\") % 2):
            raise SlashBackParserError("The amount of backslashes isn't even")
        index = 0
        found = text.find(b"\\", index)
        second_found = text.find(b"\\", found + 1)

        out = []

        current_text = None

        current_config = TextConfig()
        
        while found != -1:
            current_command = text[found + 1 : second_found]
            
            if not current_command:
                out.append(Text(text[index : found + 1], TextConfig(current_config)))
            else:
                out.append(Text(text[index : found], TextConfig(current_config)))
                current_config = self.parse_single(current_command, current_config)
            
            index = second_found + 1
            found = text.find(b"\\", index)
            second_found = text.find(b"\\", found + 1)

            if (second_found == -1):
                raise SlashBackParserError("Error while parsing. Inconsistent amount of backslashes")
            
        out.append(Text(text[index :], TextConfig(current_config)))

        return ParsedText(out)
        
open = lambda path: SlashBackParser(path)

if __name__ == "__main__":
    print(__welcome__)
    
    has_processed_files = False
    
    if len(sys.argv) != 1:
        print("")
        for filename in sys.argv[1:]:
            if filename[0] == "-":
                print(f"Invalid option: '{filename}' (ignoring)", file=sys.stderr)
            elif not os.path.exists(filename):
                print(f"File not found: '{filename}' (ignoring)", file=sys.stderr)
            else:
                has_processed_files = True
                out_filename = os.path.splitext(filename)[0] + ".md"
                processing_string = f"Processing file {filename} and saving to {out_filename}..."
                print(processing_string, end = "")
                try:
                    sbf = open(filename)
                    sbf.write(out_filename)
                    print(f"\r{processing_string} Done.")
                except SlashBackParserError as exception:
                    print(f"\nERROR: {exception!s}", file=sys.stderr)
                except Exception as exception:
                    print("\nUNHANDLED EXCEPTION OCCURED!", file=sys.stderr)
                    traceback.print_exc()
                
    if not has_processed_files:
        print("\n" + __usage__, file=sys.stderr)



import re

import os

URL_PATTERN = re.compile("\\[[^\\]]+\\]\\(([^\\)]+)\\)")

for filename in os.listdir("../"):
    if filename.endswith(".md"):
        file_ = open(os.path.join("../", filename), "r")
        content = file_.read()
        file_.close()
        
        index = 0
        while True:
            match = URL_PATTERN.search(content, index)
            if not match:
                break
            
            if match.group(1).lower() == "function-reference/README.md".lower():
                content = content[:match.start(1)] + "https://github.com/Zuzu-Typ/PyGLM/blob/master/wiki/function-reference/README.md" + content[match.end(1):]
            else:
                content = content[:match.start(1)] + match.group(1).replace(".md", "") + content[match.end(1):]
            
            index = match.end()
            
        file_ = open(filename, "w")
        file_.write(content)
        file_.close()